/* Gestione contenuto memoria Z80 - bus RC2014
 * scritto da Michele Dalla Silvestra (2023)
 * software rilasciato con licenza GNU GPL 3
 * 
 * Questo sketch permette di leggere (e modificare) la memoria
 * durante il funzionamento del microprocessore Z80.
 * 
 * Può anche caricare nella ROM una riga in formato Intel HEX
 *
 */

void setup() {
  Serial.begin(115200);
  Serial.println("Z80-DMA + EEPROM programming - 17/11/2023");
  Serial.println();
  free_bus();
}

int take_bus()
{
  pinMode(A13,OUTPUT); // Busreq
  digitalWrite(A13,0);
  for (int i=0; ((i<50) && (digitalRead(A11)==1)); i++) // Busack
    delay(100);
  if (digitalRead(A11)==1)
    return 0;
  pinMode(A0,OUTPUT); // IOreq
  pinMode(A1,OUTPUT); // RD
  pinMode(A2,OUTPUT); // WR
  pinMode(A3,OUTPUT); // Mreq
  digitalWrite(A0,1);
  digitalWrite(A1,1);
  digitalWrite(A2,1);
  digitalWrite(A3,1);
  return 1;
}

void free_bus()
{
  pinMode(A0,INPUT); // IOreq
  pinMode(A1,INPUT_PULLUP); // RD
  pinMode(A2,INPUT_PULLUP); // WR
  pinMode(A3,INPUT_PULLUP); // Mreq
//  pinMode(A1,INPUT); // RD
//  pinMode(A2,INPUT); // WR
//  pinMode(A3,INPUT); // Mreq
  pinMode(A13,INPUT); // Busreq
}

void set_addr(unsigned int addr)
{
  DDRA=0xff;
  DDRC=0xff;
  PORTA=addr>>8;
  PORTC=addr&0xff;
}

void free_addr()
{
  DDRA=0;
  DDRC=0;
  PORTA=0;
  PORTC=0;
}

void set_data(byte data)
{
  DDRL=0xff;
  PORTL=data;
}

void free_data()
{
  DDRL=0;
  PORTL=0;
}

byte memory_byte_read(unsigned int addr)
{
  set_addr(addr);
  delayMicroseconds(3);
  digitalWrite(A1,0); // RD
  digitalWrite(A3,0); // Mreq
  delayMicroseconds(3);
  byte d=PINL;
  digitalWrite(A1,1); // RD
  digitalWrite(A3,1); // Mreq
  free_addr();
  return d;
}

void memory_byte_write(unsigned int addr, byte dati)
{
  set_addr(addr);
  set_data(dati);
  delayMicroseconds(3);
  digitalWrite(A2,0); // WR
  digitalWrite(A3,0); // Mreq
  delayMicroseconds(3);
  digitalWrite(A2,1); // WR
  digitalWrite(A3,1); // Mreq
  free_addr();
  free_data();
}

void serial_print_byte(byte b)
{
  Serial.print(" ");
  if (b<0x10)
    Serial.print("0");
  Serial.print(b,HEX);
}

void serial_print_addr(unsigned int a)
{
  if (a<0x1000)
    Serial.print("0");
  if (a<0x100)
    Serial.print("0");
  if (a<0x10)
    Serial.print("0");
  Serial.print(a,HEX);
  Serial.print(" ");
}

void memory_read(unsigned int start, unsigned int len)
{
  Serial.print("richiesta bus ");
  if (take_bus())
  {
    Serial.println("avvio lettura");
    serial_print_addr(start);
    for (unsigned int i=0; i<len; i++)
    {
      serial_print_byte(memory_byte_read(start+i));
      if ((i&0xf)==0xf)
      {
        Serial.println();
        serial_print_addr(start+i+1);
      }
      else
      if ((i&0x7)==0x7)
        Serial.print(" ");
    }
    Serial.println();
    if ((len&0xf)!=0)
    {
      serial_print_addr(start+len);
      Serial.println();
    }
    Serial.println("fine lettura");
  }
  else
    Serial.println("bus timeout");
  free_bus();
}

void memory_write(unsigned int start, byte* buf, unsigned int len)
{
  Serial.print("richiesta bus ");
  if (take_bus())
  {
    Serial.println("avvio scrittura");
    serial_print_addr(start);
    for (unsigned int i=0; i<len; i++)
    {
      serial_print_byte(*(buf+i));
      memory_byte_write(start+i,*(buf+i));
      if ((i&0xf)==0xf)
      {
        Serial.println();
        serial_print_addr(start+i+1);
      }
      else
      if ((i&0x7)==0x7)
        Serial.print(" ");
    }
    Serial.println();
    if ((len&0xf)!=0)
    {
      serial_print_addr(start+len);
      Serial.println();
    }
    Serial.println("fine scrittura");
  }
  else
    Serial.println("bus timeout");
  free_bus();
}

void memory_fill(unsigned int start, byte b, unsigned int len)
{
  Serial.print("richiesta bus ");
  if (take_bus())
  {
    Serial.println("avvio scrittura");
    serial_print_addr(start);
    for (unsigned int i=0; i<len; i++)
    {
      serial_print_byte(b);
      memory_byte_write(start+i,b);
      if ((i&0xf)==0xf)
      {
        Serial.println();
        serial_print_addr(start+i+1);
      }
      else
      if ((i&0x7)==0x7)
        Serial.print(" ");
    }
    Serial.println();
    if ((len&0xf)!=0)
    {
      serial_print_addr(start+len);
      Serial.println();
    }
    Serial.println("fine scrittura");
  }
  else
    Serial.println("bus timeout");
  free_bus();
}

#define INIZIO 0x8000
#define CIFRE_INDIRIZZO 8

void cmdline_dma(char c)
{
  char testo[CIFRE_INDIRIZZO+1];
  byte lungh_testo=1;
  testo[0]=c;
  unsigned int indirizzo;
  int num_dati=-1;
  byte dati_memoria[64];
  while (c!='\n')
  {
    while (!Serial.available());
    c=Serial.read();
    if (isHexadecimalDigit(c) && (lungh_testo<CIFRE_INDIRIZZO))
      testo[lungh_testo++]=c;
    if (isSpace(c) && (lungh_testo>0))
    {
      testo[lungh_testo]=0;
      if (num_dati==-1)
      {
        indirizzo=strtol(testo,NULL,16);
        num_dati++;
      }
      else if (num_dati<64)
        dati_memoria[num_dati++]=strtol(testo,NULL,16);
      lungh_testo=0;
    }
  }
  if (num_dati==0)
    memory_read(indirizzo,64);
  if (num_dati>0)
    memory_write(indirizzo,dati_memoria,num_dati);
}

byte char2hex(char c)
{
  if (c>'@')
    return (c&0xf)+9;
  else
    return (c&0xf);
}

void cmdline_intelhex()
{
  byte dati[64];
  unsigned int indirizzo=0;
  char c;
  byte pos=0;
  byte numdati=0;
  byte tmp=0;
  byte nibble=0;
  byte check=0;
  byte tipo=0;
  while (!Serial.available());
  while (isHexadecimalDigit(c=Serial.read()))
  {
    if (nibble^=1)
      tmp=char2hex(c)<<4;
    else
    {
      tmp|=char2hex(c);
      if (pos==0)
        numdati=tmp;
      else
      if (pos<3)
        indirizzo=(indirizzo<<8)|tmp;
      else
      if (pos==3)
        tipo=tmp;
      else
      if (pos<(numdati+4))
        dati[pos-4]=tmp;
      check-=tmp;
      pos++;
    }
    while (!Serial.available());
  }
  if ((tipo==0) && (pos>4) && (check==0))
  {
    serial_print_addr(indirizzo);
    Serial.print("- ");
    serial_print_addr(indirizzo+numdati-1);
    Serial.println();
    pinMode(A2,OUTPUT); // WR
    pinMode(A3,OUTPUT); // Mreq
    digitalWrite(A2,1);
    digitalWrite(A3,1);
    for (unsigned int i=0; i<numdati; i++)
      memory_byte_write(indirizzo+i,dati[i]);
    free_bus();
  }
}

void loop() {
  if (Serial.available())
  {
    char c=Serial.read();
    if (isHexadecimalDigit(c))
      cmdline_dma(c);
    if (c==':')
      cmdline_intelhex();
  }
}
