/* Monitoraggio segnali e ROM per bus Z80 - RC2014
 * scritto da Michele Dalla Silvestra (2023)
 * software rilasciato con licenza GNU GPL 3
 * 
 * Trasmette sulla porta seriale (verso il pc) i principali
 * eventi di lettura/scrittura verso memoria e dispositivi,
 * in caso di richiesta di lettura dalla memoria, fornisce alcune
 * istruzioni operative.
 * 
 * array rom: contenuto fornito all'indirizzo 0x0000
 * array rom2: contenuto fornito all'indirizzo 0x0100
 */

void setup() {
  Serial.begin(9600);
  attachInterrupt(digitalPinToInterrupt(2),printbus,FALLING);  // WR
  attachInterrupt(digitalPinToInterrupt(3),int_pin_rd,CHANGE); // RD
}

/* ROM 0x0000: COPIA 0x0100 in 0x8000 ed esegue
 *
 * 21 00 01      ld hl,0x100
 * 11 00 80      ld de,0x8000
 * 01 12 00      ld bc,18
 * ED B0         ldir
 * C3 00 80      jp 0x8000
 *
 */
byte rom[]={ 0x21, 0x00, 0x01, 0x11, 0x00, 0x80, 0x01, 0x12, 0x00,
  0xed, 0xb0, 0xc3, 0x00, 0x80 };

/* ROM 0x0100: ANIMAZIONE LED SINISTRA-DESTRA
 * (viene copiata in ram e l'esecuzione viene fatta da RAM)
 * 
 * 3E 01           ld a,1
 * 06 07   ciclo1: ld b,7
 * D3 FF   ciclo2: out (255),a
 * 07              rlca
 * 10 FB           djnz ciclo2
 * 06 07           ld b,7
 * D3 FF   ciclo3: out (255),a
 * 0F              rrca
 * 10 FB           djnz ciclo3
 * 18 F0           jr ciclo1
 */
byte rom2[]={ 0x3e, 0x01, 0x06, 0x07, 0xd3, 0xff, 0x07, 0x10, 0xfb, 
  0x06, 0x07, 0xd3, 0xff, 0x0f, 0x10, 0xfb, 0x18, 0xf0 };

void int_pin_rd() {
  if (digitalRead(A1)!=0)
  {
    DDRL=0;
    PORTL=0;
  }
  else
  {
    if ((digitalRead(A3)==0) && (PINA==0) && (PINC<sizeof(rom)))
    {
      DDRL=0xff;
      PORTL=rom[PINC];
    }
    if ((digitalRead(A3)==0) && (PINA==1) && (PINC<sizeof(rom2)))
    {
      DDRL=0xff;
      PORTL=rom2[PINC];
    }
    printbus();
  }
}

void printbyte(byte b)
{
  for (byte i=0; i<8; i++,b<<=1)
    if ((b&0x80)!=0)
      Serial.print('1');
    else
      Serial.print('0');
}

void printbus() {
  Serial.print("Addr ");
  printbyte(PINA);
  Serial.print(' ');
  printbyte(PINC);
  Serial.print(" Data ");
  printbyte(PINL);
  if (digitalRead(A1)==0)
    Serial.print(" RD");
  if (digitalRead(A2)==0)
    Serial.print(" WR");
  if (digitalRead(A0)==0)
    Serial.print(" IOreq");
  if (digitalRead(A3)==0)
    Serial.print(" Mreq");
  if (digitalRead(A6)==0)
    Serial.print(" RESET");
  if (digitalRead(A7)==0)
    Serial.print(" M1");
  if (digitalRead(A8)==0)
    Serial.print(" Refresh");
  Serial.println();
}

void loop() {
}
